/***************************************************************************
 *   Copyright (C) 2007 by David Bitseff                                   *
 *   dbitsef@zipcon.net                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <QFileInfo>

#include "renderthread.h"
#include "flam3util.h"
#include "logger.h"
#include "isaac.h"

#ifndef NTHREADS
#define NTHREADS 1
#endif


// global 'static' used in callback
bool _stop_current_job = false;
bool loops_counted;  // total nloops is known
int count;   // nbatches/loops completed
int batch;   // current batch count
int batches; // nbatches
int nloops;  // how many total loops are there?
int nloops_count; // counted loops
int dx_time_sum;  // milliseconds
double est_loops; // batches * NTHREADS * samples
double avg_batch;
double est_remain;
double samples;   // temporal samples
double quality;
int oversample;
double last_vari;
int last_varn;
double _percent_finished;
stat_struct stats;

#if NTHREADS > 1
QMutex cb_lock;
#endif
QTime ptimer;
/**
 * this callback is needed to control the rendering function.  it also
 * calculates the estimated time remaining.  it's quite nasty since we
 * (and libflam3.a)	don't know how many calls there will be until the callback
 * is called at least a couple of times.
*/
int  _progress_callback(void* parameter, double vari, int varn)
{
	if (_stop_current_job)
		return 1;
#if NTHREADS > 1
	cb_lock.lock();
#endif
// 	cout << QString("%1: loops %2/%3, vari %4, varn %5")
// 			.arg(count).arg(nloops_count).arg(nloops)
// 			.arg(vari).arg(varn);

	dx_time_sum += ptimer.elapsed();
	ptimer.start();

	// the undocumented way to use this function.
	if ( varn == 0 ) // non-estimated
	{
		// use the batch counters to figure the metrics if vari == 0.0
		if (vari == 0.0 )
		{
			_percent_finished = (double)batch / (est_loops)* 100.;
			avg_batch = (double)dx_time_sum / batch ;
			est_remain = avg_batch * ( est_loops - batch );
			batch++;
		}
		else // varn == 0, vari != 0.0
		{
			if (last_varn == 0 && last_vari == 0.0 && !loops_counted)
			{	// figure out how many loops there will be
				nloops = (int)((qRound( (1./vari) + 0.5 ) * (est_loops)));
			}
			// use the loop counters if vari != 0.0
			_percent_finished = ((double)nloops_count/nloops )* 100.0;
			avg_batch = (double)dx_time_sum / nloops_count ;
			est_remain = avg_batch * ( nloops - nloops_count );
		}
	}
	else // varn == 1, density est. loop
	{
		if (!loops_counted) // try to count loops once
			// density est. loop when varn == 1, update the total loops once
			if ( (last_varn == 0 && vari != 0.0)
				|| (last_varn == 1 && last_vari == 0.0))
			{
				nloops += (int)((1.0/vari) * batches) ;
				if (batches > 1)
					nloops += (int)(est_loops);
				if (samples > 1 && batches == 1
								&& (quality/samples <= 4.41)) // magic?
					nloops += (int)samples*NTHREADS;
				loops_counted = true;
			}

		// use the loop counters for metrics
		_percent_finished = ((double)nloops_count/nloops )* 100.0;
		avg_batch = (double)dx_time_sum / nloops_count ;
		est_remain = avg_batch * ( nloops - nloops_count );
	}
// 	cout << QString(", %1%, avg %2, est %3, batches %4/%5")
// 			.arg(_percent_finished).arg(avg_batch).arg(est_remain)
// 			.arg(batch).arg( est_loops ) << endl;
// 	count++;
	nloops_count++;
	last_vari = vari;
	last_varn = varn;
#if NTHREADS > 1
	cb_lock.unlock();
#endif

	return 0;
}

void RenderThread::init_status_cb()
{
	nloops = 1;
	nloops_count = 1;
	dx_time_sum = 0;
	avg_batch = 0;
	est_remain = 0;
	est_loops = batches * NTHREADS  * samples;
	batch = 1;
	last_vari = -1.;
	loops_counted = false;
	stats.badvals = 0.0;
	stats.num_iters = 0;
	stats.render_seconds = 0;
// 	count  = 1;
// 	cout <<"\n";
}

RenderThread::RenderThread(flam3_frame* f)
	: running(true)
{
	// stuff to control/monitor the flam3_render function
	_stop_current_job = false;
	render_loop_flag  = false;

	flame = *f;
	flam3_init_frame(&flame);
	flame.progress = &_progress_callback;
	flame.bits = 64;
	flame.ngenomes = 1; // only render one genome
	flame.time = 0.0;
	flame.pixel_aspect_ratio = 1.0;
	flame.nthreads = NTHREADS;
	flame.bits = 64;
	flame.verbose = QString(getenv("flam3_verbose")).toInt();

	count   = 0;
	batches = 0;
	preview_request = 0;
	image_request = 0;

	so = new StatusObserver(this);
	so->start();
}

RenderThread::~RenderThread()
{
	running = false;
	so->running = false;
	delete so;
}

void RenderThread::run()
{
	logInfo("RenderThread::run : starting thread");
    while (running)
    {
		RenderRequest* job;
		rqueue_mutex.lock();
		if (preview_request != 0)
		{
			job = preview_request;
			preview_request = 0;
		}
		else if (image_request != 0)
		{
			job = image_request;
			image_request = 0;
		}
		else if (request_queue.isEmpty())
		{
			// sleep only after checking for requests
			rqueue_mutex.unlock();
			usleep(100);
			continue;
		}
		else
			job = request_queue.dequeue();
		rqueue_mutex.unlock();

		logFiner(QString("RenderThread::run : rendering request 0x%1")
				.arg((long)job,0,16));

		// make sure there is something to calculate
		bool no_pos_xf = true;
		for (flam3_xform* xf = job->genome()->xform ;
				   xf < job->genome()->xform + job->genome()->num_xforms ; xf++)
			if (xf->density > 0.0)
			{
				no_pos_xf = false;
				break;
			}
		if (no_pos_xf)
			continue;

		render_loop_flag = true;
		flam3_frame& f = flame;
		flam3_genome genome;
		f.genomes = &genome;
		genome = *(job->genome());
		genome.xform = new flam3_xform[genome.num_xforms];
		memcpy(genome.xform, job->genome()->xform,
				genome.num_xforms * sizeof(flam3_xform));

		QString rtype(job->name());
		if (job->type() != RenderRequest::File)
		{
			// scale images and previews
			int width  = genome.width;
			int height = genome.height;
			double zoom = genome.zoom;
			genome.width  = job->size().width();
			genome.height = job->size().height();
			// this call will 'sort of' move the camera to the same view
			// as a larger (or smaller) image
			genome.zoom = zoom - logf((double)width / genome.width) / 0.693147 ;
			logFiner(QString("RenderThread::run : size %1 %2  g.size %3 %4  zoom %5  g.zoom %6  %7")
				.arg(width).arg(height)
				.arg(genome.width).arg(genome.height)
				.arg(zoom).arg(genome.zoom).arg(rtype));
		}
		else
			rtype = QFileInfo(job->name()).fileName();
		msg.clear();

		int channels = 3;
		int msize = channels * genome.width * genome.height;
		unsigned char* out = new unsigned char[msize];
		unsigned char* head = out;

		// add symmetry xforms before rendering
		if (genome.symmetry != 1)
			flam3_add_symmetry(&genome, genome.symmetry);
		batches = genome.nbatches;
		samples = genome.ntemporal_samples;
		oversample = genome.spatial_oversample;
		quality = genome.sample_density;
		init_status_cb();

		logFinest(QString("allocated %1 bytes, rendering...").arg(msize));
		QTime timer;
		timer.start();
		ptimer.restart();
		rendering = true;
		flam3_render(&f, out, genome.width, 0, channels, 0, &stats);
		int millis = timer.elapsed();
		rendering = false;

		if (_stop_current_job) // if stopRendering() is called
		{
			_stop_current_job = false;
			msg = rtype.append(" rendering stopped");
			logInfo(QString("RenderThread::run : ") + msg);
			delete[] head;
			delete[] genome.xform;
			render_loop_flag = false;
			preview_request = 0;
			image_request = 0;
			request_queue.clear();
			emit flameRenderingKilled();
			continue;
		}

		QString t_format;
		if (millis > 1000 * 60 * 60)
			t_format = QString("hh:mm:ss.z");
		else if (millis > 1000 * 10)
			t_format = QString("mm:ss");
		else
			t_format = QString("ss.z");
		msg = rtype.append(" rendered in ");

		timer = QTime();
		timer = timer.addMSecs(millis);
		if (millis < 60000)
			msg += timer.toString(t_format) + QString(" seconds");
		else
			msg += timer.toString(t_format);

		logFiner(QString("RenderThread::run : ") +  msg );
		QSize buf_size(genome.width, genome.height);
		if (buf_size != img_buf.size())
			img_buf = QImage(buf_size, QImage::Format_RGB32);

		for (int h = 0 ; h < genome.height ; h++)
			for (int w = 0 ; w < genome.width ; w++, out += channels)
				img_buf.setPixel(QPoint(w, h), qRgb(out[0], out[1], out[2]));

		delete[] head;
		delete[] f.genomes->xform;

		if (job->type() == RenderRequest::File)
			img_buf.save(job->name(), "png", 100);

		job->setImage(img_buf);
		job->setFinished(true);

		// look for a free event
		RenderEvent* event = 0;
		foreach (RenderEvent* e, event_list)
			if (e->accepted())
			{
				e->accept(false);
				event = e;
				break;
			}

		if (!event)
		{
			logFinest(QString("RenderThread::run : adding event"));
			event = new RenderEvent();
			event->accept(false);
			event_list.append(event);
		}
		logFiner(QString("RenderThread::run : event list size %1")
				.arg(event_list.size()));

		event->setRequest(job);
		render_loop_flag = false;
		emit flameRendered(event);
		logFiner(QString("RenderThread::run : finished"));
	}

	logInfo("RenderThread::run : thread exiting");
}


QString& RenderThread::getMessage()
{
	if (rendering)
	{
		QTime etime = QTime();
		etime = etime.addMSecs((int)est_remain);
		QString t_format;
		if (est_remain > 1000  * 60 * 60)
			t_format = QString("hh:mm:ss");
		else if (est_remain > 1000 * 10)
			t_format = QString("mm:ss");
		else
			t_format = QString("ss.z");

		msg = QString("rendering... ")
				.append(QString::number(_percent_finished, 'f', 1))
				.append("%  ( ").append(etime.toString(t_format))
				.append(" remaining )");
	}

	return msg;
}




bool RenderThread::isRendering()
{
	return render_loop_flag;
}


void RenderThread::stopRendering()
{
	_stop_current_job = true;
}

double RenderThread::finished()
{
	return _percent_finished;
}


void RenderThread::addStatusWatcher(StatusWatcher* w)
{
	so->status.append(w);
}

void RenderThread::removeStatusWatcher(StatusWatcher* w)
{
	so->status.removeAll(w);
}



void RenderThread::render(RenderRequest* req)
{
	logFine(QString("RenderThread::render : req 0x%1").arg((long)req,0,16));
	rqueue_mutex.lock();
	if (req->type() == RenderRequest::Preview)
		preview_request = req;
	else if (req->type() == RenderRequest::Image)
		image_request = req;
	else if (request_queue.contains(req))
		logWarn(QString("RenderThread::render : req 0x%1 already queued")
				.arg((long)req,0,16));
	else
	{
		req->setFinished(false);
		request_queue.enqueue(req);
		logFine(QString("RenderThread::render : queueing req"));
	}
	rqueue_mutex.unlock();
}



// Observer used by renderthread which notifies StatusWatchers
StatusObserver::StatusObserver(RenderThread* t)
	: running(true), thread(t)
{
}

void StatusObserver::run()
{
	logInfo("StatusObserver::run : starting");
	while (running)
	{
		foreach ( StatusWatcher* w , status )
			w->setStatusMessage(thread->getMessage());
		usleep(250000);
	}
}

StatusWatcher::~StatusWatcher()
{
}


// rendering events
RenderEvent::RenderEvent()
{
	m_request = 0;
}


RenderRequest* RenderEvent::request() const
{
	return m_request;
}


void RenderEvent::setRequest(RenderRequest* r)
{
	m_request = r;
}

void RenderEvent::accept(bool t)
{
	m_accepted = t;
}

bool RenderEvent::accepted() const
{
	return m_accepted;
}



// rendering requests
RenderRequest::RenderRequest(flam3_genome* g, QSize s, QString n, Type t)
	: m_genome(g), m_size(s), m_name(n), m_type(t),
			   m_image(s, QImage::Format_RGB32), m_finished(true)
{
}


flam3_genome* RenderRequest::genome() const
{
	return m_genome;
}


void RenderRequest::setGenome(flam3_genome* value)
{
	m_genome = value;
}


QSize RenderRequest::size() const
{
	return m_size;
}


void RenderRequest::setSize(const QSize& value)
{
	m_size = value;
}


QImage& RenderRequest::image()
{
	QMutexLocker locker(&m_img_mutex);
	return m_image;
}

void RenderRequest::setImage(QImage& p)
{
	QMutexLocker locker(&m_img_mutex);
	m_image = p;
}

QString RenderRequest::name() const
{
	return m_name;
}


void RenderRequest::setName(const QString& value)
{
	m_name = value;
}


RenderRequest::Type RenderRequest::type() const
{
	return m_type;
}


void RenderRequest::setType(const Type& value)
{
	m_type = value;
}


bool RenderRequest::finished() const
{
	return m_finished;
}


void RenderRequest::setFinished(bool value)
{
	m_finished = value;
}
